/**
 * OW2 Util
 * Copyright (C) 2008 SERLI
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: UrlResourceFetcher.java 5077 2009-07-29 14:09:04Z fornacif $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.plan.fetcher.impl.url;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

import org.ow2.util.file.FileUtils;
import org.ow2.util.file.FileUtilsException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.plan.bindings.deploymentplan.DeploymentHelper;
import org.ow2.util.plan.bindings.deploymentplan.url.UrlDeployment;
import org.ow2.util.plan.bindings.exceptions.InvalidDeploymentException;
import org.ow2.util.plan.bindings.exceptions.InvalidRepositoryException;
import org.ow2.util.plan.bindings.repository.Repository;
import org.ow2.util.plan.bindings.repository.RepositoryKind;
import org.ow2.util.plan.fetcher.api.IResourceFetcher;
import org.ow2.util.plan.fetcher.api.exceptions.FetcherException;
import org.ow2.util.plan.fetcher.api.exceptions.ResourceFetcherNotResolvedException;
import org.ow2.util.plan.fetcher.api.exceptions.ResourceNotFoundException;
import org.ow2.util.plan.fetcher.impl.AbsResourceFetcherImpl;
import org.ow2.util.url.URLUtils;


/**
 * Implementation of the IResourceFetcher interface for URL deployments.
 * @author Mickaël LEDUQUE
 * @author Francois Fornaciari
 */
public class UrlResourceFetcher extends AbsResourceFetcherImpl implements IResourceFetcher {

    /**
     * The logger.
     */
    private Log logger = LogFactory.getLog(UrlResourceFetcher.class);

    /**
     * The repository where the resource has been found.
     */
    private Repository foundRepository = null;

    /**
     * The rURL to the resource.
     */
    private URL foundURL = null;

    /**
     * The local copy of the resource.
     */
    private File localFile = null;

    /**
     * Last modification time known.
     */
    private long lastModified;

    /**
     * {@inheritDoc}
     */
    public void resolve() throws FetcherException {

        if (!(this.deployment instanceof UrlDeployment)) {
            throw new FetcherException(new InvalidDeploymentException("Incompatible class " + this.deployment.getClass()));
        }
        UrlDeployment urlDeployment = (UrlDeployment) this.deployment;

        logger.debug("trying to resolve resource {0}", urlDeployment.getResource());

        String repositoryRef = null;
        try {
            repositoryRef = DeploymentHelper.getRepositoryRef(urlDeployment);
        } catch (InvalidDeploymentException e) {
            logger.error(e.getMessage());
        }

        // Try to get the resource from each repository in the repository manager
        localFile = null;
        if (repositoryRef != null) {
            Repository repository = getRepositoryManager().getRepositoryById(repositoryRef);
            if (repository == null) {
                throw new FetcherException(new InvalidRepositoryException("Undefined repository " + repositoryRef));
            }
            searchRepository(repository);
        } else {
            for (Repository repository : this.getRepositoryManager()) {
                try {
                    searchRepository(repository);
                } catch (FetcherException e) {
                    logger.debug("resource {0} was not found on repository {1}", this.deployment, repository);
                }
                if (localFile != null && localFile.exists()) {
                    // the resource was found, we can put an end to the search.
                    break;
                }
            }
        }
        if (localFile == null  || !localFile.exists()) {
            throw new FetcherException(new ResourceNotFoundException(urlDeployment.toString()));
        }
    }

    /**
     * Tries to find the requested deployment on the given repository.
     * @param repository the repository on which the search will be done.
     * @throws MalformedURLException if the URL of the repository is invalid.
     * @throws ResourceNotFoundException if the resource id not a the given URL.
     */
    private void searchRepository(final Repository repository) throws FetcherException {
        URL repositoryURL = null;
        try {
            repositoryURL = new URL(repository.getUrl());
        } catch (MalformedURLException e) {
            throw new FetcherException(new InvalidRepositoryException("Invalid URL", e));
        }

        if (repositoryURL == null) {
            logger.debug("Invalid repository URL:" + repository);
            return;
        }

        try {
            foundURL = new URL(repositoryURL + "/" + ((UrlDeployment) this.deployment).getResource());
        } catch (MalformedURLException e) {
            logger.debug("Invalid resource URL:" + e.getMessage());
        }

        // found - set the repository
        this.foundRepository = repository;

        // If the repository is local, try to return the local file
        if ("file".equals(repositoryURL.getProtocol())) {
            localFile = URLUtils.urlToFile(foundURL);
        } else {
            // create a copy of the resource.
            localFile = makeLocalFilePath(this.localRepositoriesBaseDir, this.foundRepository.getId(),
                    ((UrlDeployment) this.deployment).getResource());
        }

        // Keep the last modification time
        lastModified = localFile.lastModified();

        if ("http".equals(repositoryURL.getProtocol())) {
            // Create the local file path.
            if (!localFile.exists() || hasChanged()) {
                InputStream inputStream = null;

                try {
                    URLConnection connection = foundURL.openConnection();
                    connection.setReadTimeout(TIMEOUT);
                    inputStream = connection.getInputStream();

                    if (!localFile.getParentFile().exists()) {
                        localFile.getParentFile().mkdirs();
                    }

                    // Create the local file
                    logger.info("Downloading {0} ({1}Ko)", localFile.getName(), connection.getContentLength() / 1000);
                    FileUtils.dump(inputStream, localFile);

                    // The local file has the same last modified date than the remote resource.
                    // This is useful to check if the remote resource has been modified since the last copy.
                    localFile.setLastModified(connection.getLastModified());
                } catch (IOException ioe) {
                    logger.error("Exception while opening connection to resource {0}", foundURL);
                    throw new FetcherException("Exception while opening connection", ioe);
                } catch (FileUtilsException e) {
                    logger.error("Exception while copying data from resource {0};exception: {1}", foundURL, e);
                    throw new FetcherException("Exception while copying resource data", e);
                } finally {
                    if (inputStream != null) {
                        try {
                            // Close input stream
                            inputStream.close();
                        } catch (IOException e) {
                            // Nothing to do
                        }
                    }
                }
            }
        }

        // Keep the last modification time
        lastModified = localFile.lastModified();
    }

    /**
     * {@inheritDoc}
     */
    public RepositoryKind getSupportedRepositoryType() {
        return RepositoryKind.URL;
    }

    /**
     * {@inheritDoc}
     */
    public File getResource() throws ResourceFetcherNotResolvedException {
        if (this.localFile == null) {
            throw new ResourceFetcherNotResolvedException();
        }
        return this.localFile;
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasChanged() throws FetcherException {
        URLConnection connection = null;
        try {
            connection = foundURL.openConnection();
            connection.setReadTimeout(TIMEOUT);
            if (connection.getLastModified() != lastModified) {
                return true;
            }
        } catch (IOException e) {
            logger.debug("Cannot retrieve the last modification date for the URL {0} : {1}", foundURL, e.getMessage());
        } finally {
            try {
                if (connection.getInputStream() != null) {
                    // Close input stream
                    connection.getInputStream().close();
                }
            } catch (IOException e) {
                // Nothing to do
            }
        }

        return false;
    }

    /**
     * {@inheritDoc}
     */
    public void update() throws FetcherException {
        searchRepository(this.foundRepository);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(super.toString());
        sb.append(":useRepository=");
        if (this.foundRepository != null) {
            sb.append(this.foundRepository);
        } else {
            sb.append("null");
        }
        sb.append(":resolved=");
        sb.append(localFile != null);
        sb.append(":lastModified=");
        sb.append(this.localFile.lastModified());
        return sb.toString();
    }
}
